/* Copyright 2003-2005 Wang, Chun-Pin All rights reserved. */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <string.h>
#include <sys/param.h>
#include <sys/time.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <errno.h>
#include <glob.h>
#include <stdlib.h>

#include "ssl/ssl_port_ftpd.h"
#ifdef	USE_SSL
#include "ssl/sslapp.h"
#endif

#include "smbftpd.h"
#include "pathnames.h"
#include "param.h"

extern SYS_CONF gSysConf;
extern SESSION gSession;

/* stat a file or directory and send its information to client */
static void SMBFtpdStat(char *szName, FILE * pClient, int Verbose)
{
	struct stat statBuf;
	struct tm tmFileTime;
	char szPerm[11], szTimeStr[13];
	char szLinkTarget[MAXPATHLEN + 5], szTemp[MAXPATHLEN + 1];
#ifdef USE_SSL
	char szBuf[1024];
#endif
	int err;
	time_t t;

	lstat(szName, (struct stat *) &statBuf);

	if (S_ISLNK(statBuf.st_mode)) {
		if (gSysConf.ShowSymlinks == 0) {
			return;
		}
		strcpy(szPerm, "lrwxrwxrwx");
		err = readlink(szName, szTemp, sizeof(szTemp) - 1);
		if (err < 0) {
			syslog(LOG_ERR, "%s (%d) FTP readlink() error, error code:%d", 
				   __FILE__, __LINE__, errno);
			return;
		} else {
			szTemp[err] = '\0';
		}

		snprintf(szLinkTarget, sizeof(szLinkTarget), " -> %s", szTemp);
	} else {
		strcpy(szPerm, "----------");
		if (S_ISDIR(statBuf.st_mode))
			szPerm[0] = 'd';
		if (statBuf.st_mode & S_IRUSR)
			szPerm[1] = 'r';
		if (statBuf.st_mode & S_IWUSR)
			szPerm[2] = 'w';
		if (statBuf.st_mode & S_IXUSR)
			szPerm[3] = 'x';
		if (statBuf.st_mode & S_IRGRP)
			szPerm[4] = 'r';
		if (statBuf.st_mode & S_IWGRP)
			szPerm[5] = 'w';
		if (statBuf.st_mode & S_IXGRP)
			szPerm[6] = 'x';
		if (statBuf.st_mode & S_IROTH)
			szPerm[7] = 'r';
		if (statBuf.st_mode & S_IWOTH)
			szPerm[8] = 'w';
		if (statBuf.st_mode & S_IXOTH)
			szPerm[9] = 'x';
		szLinkTarget[0] = '\0';		
	} 
	memcpy(&tmFileTime, localtime(&(statBuf.st_mtime)), sizeof(struct tm));
	time(&t);
	if (tmFileTime.tm_year == localtime(&t)->tm_year){
		strncpy(szTimeStr, ctime(&(statBuf.st_mtime)) + 4, 12);
	} else {
		strftime(szTimeStr, sizeof(szTimeStr), "%b %d  %G", &tmFileTime);
	}
	szTimeStr[12]='\0';

	if (Verbose) {
#ifdef USE_SSL
		if (ssl_data_active_flag) {
			snprintf(szBuf, sizeof(szBuf), "%s %3s %-8s %-8s %12llu %s %s%s\r\n", 
					 szPerm, "1", user_from_uid(statBuf.st_uid, 0), group_from_gid(statBuf.st_gid, 0),
					 (unsigned long long) statBuf.st_size,
					 szTimeStr, szName, szLinkTarget);
			ssl_write(ssl_data_con, szBuf, strlen(szBuf));
		} else
#endif
		fprintf(pClient, "%s %3s %-8s %-8s %12llu %s %s%s\r\n", szPerm, "1", 
				user_from_uid(statBuf.st_uid, 0), group_from_gid(statBuf.st_gid, 0),
				(unsigned long long) statBuf.st_size,
				szTimeStr, szName, szLinkTarget);
	} else {
#ifdef USE_SSL
		if (ssl_data_active_flag) {
			snprintf(szBuf, sizeof(szBuf), "%s\r\n", szName);
			ssl_write(ssl_data_con, szBuf, strlen(szBuf));
		} else
#endif
		fprintf(pClient,"%s\r\n", szName);
	}

	return;
}

/* List share's information and send to client */
static void SMBFtpdRootStat(char *szShare, char *szName, FILE * pClient, int Verbose)
{
	struct stat statBuf;
	struct tm tmFileTime;
	char szPerm[11], szTimeStr[13];
#ifdef USE_SSL
	char szBuf[1024];
#endif
	
	time_t t;

	lstat(szName, (struct stat *) &statBuf);
	strncpy(szPerm, "dr-x------", sizeof(szPerm));
	memcpy(&tmFileTime, localtime(&(statBuf.st_mtime)), sizeof(struct tm));
	time(&t);
	if (tmFileTime.tm_year == localtime(&t)->tm_year){
		strncpy(szTimeStr, ctime(&(statBuf.st_mtime)) + 4, 12);
	} else {
		strftime(szTimeStr, sizeof(szTimeStr), "%b %d  %G", &tmFileTime);
	}
	szTimeStr[12]='\0';
	if (Verbose) {
#ifdef USE_SSL
		if (ssl_data_active_flag) {
			snprintf(szBuf, sizeof(szBuf), "%s %3i %s %-5s %12llu %s %s\r\n", 
					 szPerm , 1, "root", "users", (unsigned long long) statBuf.st_size, 
					 szTimeStr, szShare);
			ssl_write(ssl_data_con, szBuf, strlen(szBuf));
		} else
#endif
		fprintf(pClient, "%s %3i %s %-5s %12llu %s %s\r\n", szPerm , 1, "root", "users", 
				(unsigned long long) statBuf.st_size, szTimeStr, szShare);
	}else{
#ifdef USE_SSL
		if (ssl_data_active_flag) {
			snprintf(szBuf, sizeof(szBuf), "%s\r\n", szShare);
			ssl_write(ssl_data_con, szBuf, strlen(szBuf));
		} else
#endif
		fprintf(pClient, "%s\r\n", szShare);
	}

	return;
}

char gszMyCWD[MAXPATHLEN];
/* Do LIST command */
void SMBDirList(char *szPath, FILE * pClient, int Verbose, int Recursive)
{
	struct stat statBuf;
	char szCwd[MAXPATHLEN+1], szRealPath[MAXPATHLEN+1], szSearchPath[MAXPATHLEN+1];
	char szBuf[MAXPATHLEN], szDir[MAXPATHLEN+1], *szName, *szPtr;
	int i, Flags = FLAG_NO_FOLLOW_LAST_LINK, NeedConcatePath = 0;
	glob_t globBuf;

	if ((strstr(szPath, "/.")) && strchr(szPath, '*'))
		return; /* DoS protection */

	// When NLIST, we have to show the path so the mget can work
	if (!Verbose && strchr(szPath, '/')) {
		NeedConcatePath = 1;
	}

	snprintf(szSearchPath, sizeof(szSearchPath), "%s", szPath);
	if (strchr(szSearchPath, '*') || strchr(szSearchPath, '?') || strchr(szSearchPath, '[')) {
		szName = strrchr(szSearchPath, '/');
		if (NULL != szName) {
			*szName = '\0';
			szName++;
			if (strchr(szSearchPath, '*') || strchr(szSearchPath, '?') || strchr(szSearchPath, '[')) {
				// We don't support xxx/*/xxx
				return;
			}
			Flags = 0;
		} else {
			strcpy(szSearchPath, ".");
			szName = szPath;
		}
	} else {
		szName = NULL;
		if (szSearchPath[strlen(szSearchPath) - 1] == '/') {
			Flags = 0;
		}
	}
	
	szPtr = SzGetRealPath(gSession.pValidShare, szSearchPath, Flags);
	if (NULL == szPtr) {
		goto ErrorOut;
	}
	snprintf(szRealPath, sizeof(szRealPath), "%s", szPtr);

	if ((gSession.DefaultMode == MODE_SMB) && strcmp(szRealPath, _PATH_SMB_FTPD_ROOT) == 0) {
		struct SHARE_SET *pSet = gSession.pValidShare;
		while (pSet) {
			// List all readable shares
			if (!(pSet->flags & ATTRIB_NO_BROWSE)) {

				SMBFtpdRootStat(pSet->szShare,
								pSet->szPath, 
								pClient, Verbose);
			}
			pSet = pSet->pNext;
		}
	} else {
		if (CheckShareFlag(gSession.pValidShare, szRealPath, ATTRIB_NO_LIST)) {
			return;
		}

		getcwd(szCwd, sizeof(szCwd) - 1);

		if (lstat(szRealPath, (struct stat *) &statBuf) < 0) {
			goto ErrorOut;
		} else  if (S_ISDIR(statBuf.st_mode)) {
			chdir(szRealPath);
			if (szName) {
				glob(szName, 0, NULL, &globBuf);
			} else {
				// Also list file begin with dot
				glob(".*", 0, NULL, &globBuf);
				glob("*", GLOB_APPEND, NULL, &globBuf);
			}
		} else {
			glob(szRealPath, 0, NULL, &globBuf);
		}
				
		// The recursive stuff is for ffftp, it recursive lists all files
		// before deletes them.
		if (Recursive == 1) {
			snprintf(gszMyCWD, sizeof(gszMyCWD), "%s", szCwd);
		} else if (Recursive > 1){

			if (strncmp(gszMyCWD, szRealPath, strlen(gszMyCWD)) == 0) {
				szPtr = szRealPath + strlen(gszMyCWD) + 1;
			} else {
				szPtr = szRealPath;
			}
			
			snprintf(szBuf, sizeof(szBuf), "\r\n./%s:\r\n", szPtr);

#ifdef USE_SSL
			if (ssl_data_active_flag) {
				ssl_write(ssl_data_con, szBuf, strlen(szBuf));
			} else 
#endif
			fprintf(pClient, szBuf);
		}


		if (NeedConcatePath) {
			chdir(szCwd);
			strcpy(szDir, szPath);
			szPtr = strrchr(szDir, '/');
			*szPtr = '\0';
		}
		for (i = 0; i < globBuf.gl_pathc; i++){
			if ((strcmp(globBuf.gl_pathv[i],".")==0) ||
				(strcmp(globBuf.gl_pathv[i],"..")==0)) {
				continue;
			}
			if (NeedConcatePath) {
                snprintf(szBuf, sizeof(szBuf), "%s/%s", szDir, globBuf.gl_pathv[i]);
				SMBFtpdStat(szBuf, pClient, Verbose);
			} else {
				SMBFtpdStat(globBuf.gl_pathv[i], pClient, Verbose);
			}
		}

		if (Recursive) {
			for (i = 0; i < globBuf.gl_pathc; i++) {

				if ((strcmp(globBuf.gl_pathv[i],".")==0) ||
					(strcmp(globBuf.gl_pathv[i],"..")==0)) {
					continue;
				}
				bzero(&statBuf, sizeof(statBuf));
				lstat(globBuf.gl_pathv[i], (struct stat *) &statBuf);
				if (S_ISDIR(statBuf.st_mode)) {
					SMBDirList(globBuf.gl_pathv[i], pClient, Verbose, Recursive+1);
				}
			}
		}
		chdir(szCwd);
		globfree(&globBuf);
	}

	return;

	ErrorOut:
	#ifdef USE_SSL
	if (ssl_data_active_flag) {
		snprintf(szBuf, sizeof(szBuf), "ftpd: %s: No such file or directory.\r\n", szPath);
		ssl_write(ssl_data_con, szBuf, strlen(szBuf));
	} else 
	#endif
	fprintf(pClient,"ftpd: %s: No such file or directory.\r\n", szPath);
	return;
}
