/* Copyright 2003-2005 Wang, Chun-Pin All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <grp.h>
#include <pwd.h>

#include "misc.h"

/* This Function will remove the spaces ' ' and tab '\t' and newline '\n'
 * in the front and tail of the given szString.
 * For example, if the string is "   this is a string  ".
 * it would become "this is a string".
 *
 * Return value:
 *        *szString
 */
char *SzTrimSpace(char *szString)
{
	char *szPtr1, *szPtr2;

	if (NULL == szString) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return NULL;
	}
	
	/* Skip the leading spaces */
	for (szPtr1 = szString; (((*szPtr1 == ' ') || 
							  (*szPtr1 == '\t') || (*szPtr1 == '\n')) && 
							 (*szPtr1 != '\0')); szPtr1++);

	/* Remove the leading spaces */
	for (szPtr2 = szString; *szPtr2 != '\0'; szPtr2++, szPtr1++) {
		*szPtr2 = *szPtr1;
	}

	/* Look for ending spaces */
	szPtr2 = NULL;
	for (szPtr1 = szString; *szPtr1 != '\0'; szPtr1++) {
		if ((*szPtr1 == ' ') || (*szPtr1 == '\t') || (*szPtr1 == '\n')) {
			if (szPtr2 == NULL) {
				szPtr2 = szPtr1;
			}
		} else {
			szPtr2 = NULL;
		}
	}
	if (szPtr2 != NULL) {
		*szPtr2 = '\0';
	}

	return szString;
}

/* This Function will remove the spaces ' ', tab '\t', '"', '\'' and newline '\n'
 * in the front and tail of the given szString.
 * For example, if the string is "   this is a string  ".
 * it would become "this is a string".
 *
 * Return value:
 *        *szString
 */
char *SzTrimString(char *szString)
{
	char *szPtr1, *szPtr2;

	if (NULL == szString) {
		syslog(LOG_ERR, "%s (%d) bad parameter", __FILE__, __LINE__);
		return NULL;
	}
	
	/* Skip the leading spaces */
	for (szPtr1 = szString; (((*szPtr1 == ' ') || (*szPtr1 == '\t') || 
							  (*szPtr1 == '\n') || (*szPtr1 == '"') || 
							  (*szPtr1 == '\'')) && 
							 (*szPtr1 != '\0')); szPtr1++);

	/* Remove the leading spaces */
	for (szPtr2 = szString; *szPtr2 != '\0'; szPtr2++, szPtr1++) {
		*szPtr2 = *szPtr1;
	}

	/* Look for ending spaces */
	szPtr2 = NULL;
	for (szPtr1 = szString; *szPtr1 != '\0'; szPtr1++) {
		if ((*szPtr1 == ' ') || (*szPtr1 == '\t') || (*szPtr1 == '\n') || 
			(*szPtr1 == '"') || (*szPtr1 == '\'')) {
			if (szPtr2 == NULL) {
				szPtr2 = szPtr1;
			}
		} else {
			szPtr2 = NULL;
		}
	}
	if (szPtr2 != NULL) {
		*szPtr2 = '\0';
	}

	return szString;
}

/* Check whether user is in szList
 * the szList is a string of users that separate by '.'.
 * For example szList = "user1, user2, user3..."
 *
 * Return Values:
 *	TRUE: yes
 *	FALSE: no
 */
BOOL IsUserInList(char *szUser, char *szList)
{
	struct group *grp = NULL;
	struct passwd *pw = NULL;
	char *szBuf = NULL, **szGrpMem, *szTok;
	int err = FALSE;

	if (!szUser || !szList) {
		return err;
	}

	pw = getpwnam(szUser);
	if (NULL == pw) {
		syslog(LOG_ERR, "%s (%d) failed to get user [%s]",
			   __FILE__, __LINE__, szUser);
		return err;
	}

	szBuf = (char *)malloc(strlen(szList) + 1);
	if (szBuf == NULL) {
		syslog(LOG_ERR, "%s (%d) failed to allocate memory, errno:%d(%s)",
			   __FILE__, __LINE__, errno, strerror(errno));
		return err;
	}
	strcpy(szBuf, szList);

	for (szTok = strtok(szBuf, ","); szTok; szTok = strtok(NULL, ",")) {
		SzTrimSpace(szTok);
		if (*szTok != '@') {	   //user
			if (0 == strcmp(szUser, szTok)) {
				err = TRUE;
				break;
			}
		} else {
			
			grp = getgrnam(szTok+1);
			if (NULL == grp) {
				syslog(LOG_ERR, "%s (%d) failed to get group [%s]",
					   __FILE__, __LINE__, szTok+1);
				continue;
			}
			if (pw->pw_gid == grp->gr_gid) {
				err = TRUE;
				goto Error;
			}
			szGrpMem = grp->gr_mem;
			while (*szGrpMem) {
				if (strcmp(*szGrpMem, szUser) == 0) {
					err = TRUE;
					goto Error;
				}		
				szGrpMem++;
			}
		}
	}

Error:
	if (szBuf != NULL) {
		free(szBuf);
	}
	return err;
}

BOOL IsUserInGroup(char *szUser, char *szGroup)
{
	struct group *grp = NULL;
	struct passwd *pw = NULL;
	char **szGrpMem;
	int err = FALSE;

	if (!szUser || !szGroup) {
		return err;
	}

	pw = getpwnam(szUser);
	if (NULL == pw) {
		syslog(LOG_ERR, "%s (%d) failed to get user [%s]",
			   __FILE__, __LINE__, szUser);
		return err;
	}

	grp = getgrnam(szGroup);
	if (NULL == grp) {
		syslog(LOG_ERR, "%s (%d) failed to get group [%s]",
			   __FILE__, __LINE__, szGroup);
		return err;
	}

	if (pw->pw_gid == grp->gr_gid) {
		return TRUE;
	}
	
	szGrpMem = grp->gr_mem;
	while (*szGrpMem) {
		if (strcmp(*szGrpMem, szUser) == 0) {
			err = TRUE;
			break;
		}		
		szGrpMem++;
	}

	return err;
}

char *szRealPath(char *szPath)
{
	static char szRealPath[PATH_MAX];
	char szCurPath[PATH_MAX];

	bzero(szRealPath, sizeof(szRealPath));
	bzero(szCurPath, sizeof(szCurPath));

	getcwd(szCurPath, sizeof(szCurPath));
	if (chdir(szPath) == 0) {
		getcwd(szRealPath, sizeof(szRealPath));
		chdir(szCurPath);
		return szRealPath;
	} else {
		return NULL;
	}
}
