/* Copyright 2003-2005 Wang, Chun-Pin All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <sys/param.h>
#include <syslog.h>
#include <unistd.h>
#include "config.h"
#include "misc.h"
#include "param.h"
#include "pathnames.h"

extern SYS_CONF gSysConf;

CONF_OPT_LIST AllConfList[] = {
	{"ChrootSet",            OPT_TYPE_SET,    &(gSysConf.ChrootSet)},
	{"MaxDownloadRate",      OPT_TYPE_SET,    &(gSysConf.MaxDownloadRate)},
	{"MaxUploadRate",        OPT_TYPE_SET,    &(gSysConf.MaxUploadRate)},
	{"ShareConfPath",        OPT_TYPE_PATH,   &(gSysConf.szShareConf)},
	{"TransferLog",          OPT_TYPE_STR,    &(gSysConf.szTransferLog)},
	{"ServerName",           OPT_TYPE_STR,    &(gSysConf.szServerName)},
	{"ListenOnAddress",      OPT_TYPE_STR,    &(gSysConf.szListenOnAddress)},
	{"Port",                 OPT_TYPE_STR,    &(gSysConf.szPort)},
	{"ExceptionList",        OPT_TYPE_STR,    &(gSysConf.szExceptionList)},
	{"PidFile",              OPT_TYPE_STR,    &(gSysConf.szPidFile)},
	{"NoLoginList",          OPT_TYPE_STR,    &(gSysConf.szNoLoginList)},
	{"DefaultMode",          OPT_TYPE_MODE,   &(gSysConf.DefaultMode)},
	{"DebugMode",            OPT_TYPE_YES_NO, &(gSysConf.DebugMode)},
	{"LogCommand",           OPT_TYPE_YES_NO, &(gSysConf.LogCommand)},
	{"DoWtmpLog",            OPT_TYPE_YES_NO, &(gSysConf.DoWtmpLog)},
	{"AnonymousLogin",       OPT_TYPE_YES_NO, &(gSysConf.AnonymousLogin)},
	{"AnonymousOnly",        OPT_TYPE_YES_NO, &(gSysConf.AnonymousOnly)},
	{"EmptyPasswdLogin",     OPT_TYPE_YES_NO, &(gSysConf.EmptyPasswdLogin)},
	{"ShowSymlinks",         OPT_TYPE_YES_NO, &(gSysConf.ShowSymlinks)},
	{"RequireValidShell",    OPT_TYPE_YES_NO, &(gSysConf.RequireValidShell)},
	{"DisableEPSV",          OPT_TYPE_YES_NO, &(gSysConf.DisableEPSV)},
	{"RestrictedPorts",      OPT_TYPE_YES_NO, &(gSysConf.RestrictedPorts)},
	{"RestrictedDataPorts",  OPT_TYPE_YES_NO, &(gSysConf.RestrictedDataPorts)},
	{"Umask",                OPT_TYPE_OCTAL,  &(gSysConf.Umask)},
	{"MaxConnection",        OPT_TYPE_INT,    &(gSysConf.MaxConnection)},
	{"TimeOut",              OPT_TYPE_INT,    &(gSysConf.TimeOut)},
	{"MaxTimeOut",           OPT_TYPE_INT,    &(gSysConf.MaxTimeOut)},
	// SSL/TLS Options
	{"SecurityPolicy",       OPT_TYPE_STR,    &(gSysConf.szSecurityPolicy)},
	{"EncryptionType",       OPT_TYPE_STR,    &(gSysConf.szEncryptionType)},
	{"NormalUserMustSecure", OPT_TYPE_YES_NO, &(gSysConf.NormalUserMustSecure)},
	{"AnonymDisableSecure",  OPT_TYPE_YES_NO, &(gSysConf.AnonymDisableSecure)},
	{"SSLCertFile",          OPT_TYPE_PATH,   &(gSysConf.szSSLCertFile)},
	{"SSLKeyFile",           OPT_TYPE_PATH,   &(gSysConf.szSSLKeyFile)},
	{"SSLDebugMode",         OPT_TYPE_YES_NO, &(gSysConf.SSLDebugMode)},
	{"SSLDebugLog",          OPT_TYPE_STR,    &(gSysConf.szSSLDebugLog)},

	{NULL, 0, NULL}
};

void ConfigDefSet()
{
	char szHostname[MAXHOSTNAMELEN];

	bzero(szHostname, sizeof(szHostname));
	if (gethostname(szHostname, sizeof(szHostname)) == 0) {
		gSysConf.szServerName = strdup(szHostname);
	} else {
		gSysConf.szServerName = strdup("SMB");
	}
	gSysConf.szPort = strdup("ftp");
	gSysConf.Umask = 022;
	gSysConf.TimeOut = 900;
	gSysConf.MaxTimeOut = 7200;
	gSysConf.RestrictedDataPorts = 1;
	gSysConf.DefaultMode = MODE_NORMAL;
	gSysConf.RestrictedPorts = 1;
	gSysConf.RestrictedDataPorts = 1;

	// SSL/TLS options
	gSysConf.szSecurityPolicy = strdup("both");
	gSysConf.szEncryptionType = strdup("both");
	gSysConf.szSSLCertFile = strdup(_PATH_SSL_CERT_FILE);
	gSysConf.szSSLKeyFile = strdup(_PATH_SSL_KEY_FILE);
}

int ConfigParse(char *szConfig)
{
	CONF_OPT_LIST *OptList;
	FILE *pfConf = NULL;
#ifdef	_HAVE_FGETLN_
	char *szLine;
#else
	char szLine[LINE_MAX];
#endif
	char *szOption, *szOptArgs;
	char *szArg1, *szArg2;
	int err = -1, len = 0, i;
	struct stat statBuf;

	if (NULL == szConfig) {
		syslog(LOG_ERR, "%s (%d) Please specify the config file", __FILE__, __LINE__);
		return -1;
	}

	pfConf = fopen(szConfig, "r");
	if (!pfConf) {
		syslog(LOG_ERR, "%s (%d) Failed to open [%s] (%s)", __FILE__, __LINE__,
			   szConfig, strerror(errno));
		return -1;
	}

#ifdef	_HAVE_FGETLN_
	while (NULL != (szLine = fgetln(pfConf, &len))) {
		szLine[len] = 0;
#else
	while (NULL != (fgets(szLine, sizeof(szLine), pfConf))) {
		len = strlen(szLine);
#endif
		/* skip comments */
		if (szLine[0] == '#')
			continue;
		
		if (szLine[len - 1] == '\n') {
			szLine[len - 1] = '\0';
		}
		szOption = szLine;
		// Trim space
		while ((*szOption == ' ') || (*szOption == '\t')) {
			szOption++;
		}
		if (szOption == '\0') {
			continue;
		}

		szOptArgs = szOption;
		while ((*szOptArgs != ' ') && (*szOptArgs != '\t') && (*szOptArgs != '\0')) {
			szOptArgs++;
		}
		if (szOptArgs == szOption) {
			// Empty line
			continue;
		}

		*szOptArgs = '\0';
		szOptArgs++;

		OptList = NULL;
		for (i = 0; i < sizeof(AllConfList); i++) {
			if (AllConfList[i].szOptString == NULL) {
				break;
			}
			if (strcasecmp(szOption, AllConfList[i].szOptString) == 0) {
				OptList = &AllConfList[i];
				break;
			}
		}
		if (!OptList) {
			syslog(LOG_ERR, "%s (%d) Skip unknown option %s", __FILE__, __LINE__, szOption);
			continue;
		}
		SzTrimString(szOptArgs);
		if (*szOptArgs == '\0') {
			syslog(LOG_ERR, "%s (%d) bad syntax of config option %s",
					   __FILE__, __LINE__, OptList->szOptString);
			goto Error;
		}

		switch (OptList->OptType) {
		case OPT_TYPE_OCTAL:
			*((int *)(OptList->SysConfMember)) = (int)strtol(szOptArgs, (char **)NULL, 8);
			if (*((int *)(OptList->SysConfMember)) <= 0) {
				syslog(LOG_ERR, "%s (%d) bad number of config option %s",
					   __FILE__, __LINE__, OptList->szOptString);
				goto Error;
			}
			break;
		case OPT_TYPE_INT:
			*((int *)(OptList->SysConfMember)) = (int)strtol(szOptArgs, (char **)NULL, 10);
			if (*((int *)(OptList->SysConfMember)) <= 0) {
				syslog(LOG_ERR, "%s (%d) bad number of config option %s",
					   __FILE__, __LINE__, OptList->szOptString);
				goto Error;
			}
			break;
		case OPT_TYPE_PATH:
			if (0 != stat(szOptArgs, &statBuf) || !S_ISREG(statBuf.st_mode)) {
				syslog(LOG_ERR, "%s (%d) The option %s (%s) does not exist",
					   __FILE__, __LINE__, OptList->szOptString, szOptArgs);
				goto Error;
			}
			/* Fall down */
		case OPT_TYPE_STR:
			*((char **)OptList->SysConfMember) = strdup(szOptArgs);
			if (*((char **)OptList->SysConfMember) == NULL) {
				syslog(LOG_ERR, "%s (%d) Ran out of memory.", __FILE__, __LINE__);
				goto Error;
			}
			break;
		case OPT_TYPE_YES_NO:
			if (strcasecmp(szOptArgs, "yes") == 0) {
				*((int *)OptList->SysConfMember) = 1;
			} else {
				*((int *)OptList->SysConfMember) = 0;
			}
			break;
		case OPT_TYPE_MODE:
			if (strcasecmp(szOptArgs, "SMB") == 0) {
				*((int *)OptList->SysConfMember) = MODE_SMB;
			} else {
				*((int *)OptList->SysConfMember) = MODE_NORMAL;
			}
			break;
		case OPT_TYPE_SET:
			szArg1 = strtok(szOptArgs, " \t");
			szArg2 = strtok(NULL, " \t");
			if (szArg1 && szArg2) {
				struct OPT_SET *p;
				p = calloc(sizeof(struct OPT_SET), 1);
				p->szKey = strdup(szArg1);
				p->szValue = strdup(szArg2);
				if (p->szKey == NULL || p->szValue == NULL) {
					syslog(LOG_ERR, "%s (%d) Ran out of memory.", __FILE__, __LINE__);
					goto Error;
				}
				p->next = *((struct OPT_SET **)OptList->SysConfMember);
				*((struct OPT_SET **)OptList->SysConfMember) = p;
			} else {
				syslog(LOG_ERR, "%s (%d) bad syntax of config option %s",
					   __FILE__, __LINE__, OptList->szOptString);
				goto Error;
			}
			break;
			
		case OPT_TYPE_UNKNOWN:
			break;
		}
	}

	if ((gSysConf.DefaultMode == MODE_SMB) || 
		((gSysConf.DefaultMode == MODE_NORMAL) && (gSysConf.szExceptionList))) {
		if (gSysConf.szShareConf == NULL) {
			syslog(LOG_ERR, "%s (%d) Enable SMB mode but failed to find samba config file.",
				   __FILE__, __LINE__);
			goto Error;
		}
	}
	if (gSysConf.TimeOut > gSysConf.MaxTimeOut) {
		syslog(LOG_ERR, "%s (%d) The MaxTimeOut should be bigger than TimeOut.",
			   __FILE__, __LINE__);
		goto Error;
	}

	err = 0;
Error:
	if (pfConf) {
		fclose(pfConf);
	}

	return err;
}


void OptSetFree(struct OPT_SET *pSet)
{
	while (pSet) {
		struct OPT_SET *p = pSet;
		pSet = p->next;
		if (p->szKey) {
			free(p->szKey);
		}
		if (p->szValue) {
			free(p->szValue);
		}
		free(p);
	}
}

void ConfigRelease()
{
	CONF_OPT_LIST *pConfList = AllConfList;

	while (pConfList->szOptString) {
		switch (pConfList->OptType) {
		case OPT_TYPE_SET:
			OptSetFree(*((struct OPT_SET **)pConfList->SysConfMember));
			*(char **)pConfList->SysConfMember = NULL;
			break;
		case OPT_TYPE_STR:
			if (*(char **)pConfList->SysConfMember != NULL) {
				free(*((char **)pConfList->SysConfMember));
				*(char **)pConfList->SysConfMember = NULL;
			}
			break;
		default:
			*(int **)pConfList->SysConfMember = 0;
			break;
		}
		pConfList++;
	}

	bzero(&gSysConf, sizeof(gSysConf));
}
