/* Copyright 2003-2005 Wang, Chun-Pin All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <grp.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>

#include "param.h"
#include "misc.h"

/* Get value from set by username.
 * The Key is set is username or @group, we will traverse the sets
 * and find key match szUser or key is the group that szUser belongs
 * to.
 *
 * Return values:
 *	NULL: not found
 *	value
 */
char *SetGetValue(struct OPT_SET *set, char *szUser)
{
	struct OPT_SET *p = set;
	char *szKey;
	int Match = 0;

	if (!szUser || !set) {
		return NULL;
	}

	while (p) {
		szKey = p->szKey;
		if (*szKey != '@') {	   //user
			if (0 == strcmp(szKey, szUser)) {
				Match = 1;
			}
		} else if (*(szKey+1) == '\0') { /* "@" match all */
			Match = 1;
		} else {	/* Group */
			if (IsUserInGroup(szUser, szKey + 1)) {
				Match = 1;
			}
		}
		if (Match) {
			return p->szValue;
		}
		p = p->next;
	}
	return NULL;
}

char *szGetChrootPath(struct OPT_SET *ChrootSet, char *szUser)
{
	return SetGetValue(ChrootSet, szUser);
}

/*
 * Get value from set by username and transfer the value into
 * bps (bits per second).
 */
off_t GetRateFromSet(struct OPT_SET *set, char *szUser)
{
	off_t size;
	char *szValue;

	szValue = SetGetValue(set, szUser);
	if (NULL == szValue) {
		return 0;
	}
	size = 1024 * strtol(szValue, (char **)NULL, 10);

	return size;
}

int GetSmbFTPMode(int DefaultMode, char *szExceptionList, char *szUser)
{
	int Mode = DefaultMode;

	if (IsUserInList(szUser, szExceptionList)) {
		if (DefaultMode == MODE_NORMAL) {
			Mode = MODE_SMB;
		} else if (DefaultMode == MODE_SMB) {
			Mode = MODE_NORMAL;
		}
	}

	return Mode;
}

/* This function is used to block/unblock signale in TransferThrottle()
 * to avoid signal received when select.
 */
static void TransferRateSigmask(int block)
{
	static sigset_t sig_set;

	if (block) {
		sigemptyset(&sig_set);

		sigaddset(&sig_set, SIGCHLD);
		sigaddset(&sig_set, SIGINT);
		sigaddset(&sig_set, SIGQUIT);
		sigaddset(&sig_set, SIGURG);
		sigaddset(&sig_set, SIGIO);
		sigaddset(&sig_set, SIGBUS);
		sigaddset(&sig_set, SIGHUP);

		while (sigprocmask(SIG_BLOCK, &sig_set, NULL) < 0) {
			if (errno == EINTR) {
				continue;
			}
			break;
		}
	} else {
		while (sigprocmask(SIG_UNBLOCK, &sig_set, NULL) < 0) {
			if (errno == EINTR) {
				continue;
			}
			break;
		}
	}
	return;
}

/* Returns the difference, in milliseconds, between the given timeval and
 * now.
 */
static long ElapsedTimeCount(struct timeval *tvSince) 
{
	struct timeval tvNow;
	gettimeofday(&tvNow, NULL);

	return (((tvNow.tv_sec - tvSince->tv_sec) * 1000L) +
			((tvNow.tv_usec - tvSince->tv_usec) / 1000L));
}

void TransferThrottle(off_t SentLeng, struct timeval *tvSince, off_t Rate) 
{
	time_t IdealTime = 0, ElapsedTime = 0;

	/* Calculate the time interval since the transfer of data started. */
	ElapsedTime = ElapsedTimeCount(tvSince);

	IdealTime = SentLeng * 1000L / Rate;

	if (IdealTime > ElapsedTime) {
		struct timeval tvDelay;

		/* Setup for the select.  We use select() instead of usleep() because it
		 * seems to be far more portable across platforms.
		 *
		 * IdealTime and ElapsedTime are in milleconds, but tv_usec will be microseconds,
		 * so be sure to convert properly.
		 */
		tvDelay.tv_usec = (IdealTime - ElapsedTime) * 1000;
		tvDelay.tv_sec = tvDelay.tv_usec / 1000000L;
		tvDelay.tv_usec = tvDelay.tv_usec % 1000000L;

		/* No interruptions, please... */
		TransferRateSigmask(TRUE);

		if (select(0, NULL, NULL, NULL, &tvDelay) < 0)
			syslog(LOG_ERR, "warning: unable to throttle bandwidth: %s",
				   strerror(errno));

		TransferRateSigmask(FALSE);

	}

	return;
}
